import os
from urllib.parse import urljoin
from django.core.files.storage import FileSystemStorage
from django.utils.encoding import filepath_to_uri
from django.utils.deconstruct import deconstructible
from django.conf import settings
from accrete.tenant import get_tenant


@deconstructible(path="django.core.files.storage.FileSystemStorage")
class TenantFileSystemStorage(FileSystemStorage):

    @property
    def base_location(self):
        tenant = get_tenant()
        base_dir = f'{settings.MEDIA_ROOT}/{tenant.id}'
        os.makedirs(os.path.dirname(base_dir), exist_ok=True)
        return base_dir

    @property
    def location(self):
        return os.path.abspath(self.base_location)

    @property
    def base_url(self):
        if self._base_url is not None and not self._base_url.endswith("/"):
            self._base_url += "/"
        res = self._value_or_setting(self._base_url, f'{settings.MEDIA_URL.strip("/")}/{get_tenant().id}')
        return res

    def url(self, name):
        if self.base_url is None:
            raise ValueError("This file is not accessible via a URL.")
        url = filepath_to_uri(name)
        if url is not None:
            url = url.lstrip("/")
        return urljoin(f'/{self.base_url}/', url)



import logging
from django import forms
from django.contrib.auth import get_user_model
from accrete.tenant import get_tenant

_logger = logging.getLogger(__name__)
User = get_user_model()


class TenantForm(forms.Form):

    def __init__(self, *args, **kwargs):
        self.tenant = get_tenant()
        super().__init__(*args, **kwargs)
        fields_to_filter = filter(
            lambda x:
            hasattr(x, 'queryset') and hasattr(x.queryset.model, 'tenant'),
            self.fields.values()
        )
        for field in fields_to_filter:
            field.queryset = field.queryset.filter(tenant=self.tenant)


class TenantModelForm(forms.ModelForm):

    def __init__(self, *args, **kwargs):
        self.tenant = get_tenant()
        super().__init__(*args, **kwargs)
        fields_to_filter = filter(
            lambda x:
            hasattr(x, 'queryset') and hasattr(x.queryset.model, 'tenant'),
            self.fields.values()
        )
        for field in fields_to_filter:
            field.queryset = field.queryset.filter(tenant=self.tenant)

    def save(self, commit=True):
        super().save(commit=False)
        if not self.instance.pk or not self.instance.tenant:
            self.instance.tenant = self.tenant

        if commit:
            self.instance.save()
            self.save_m2m()
        return self.instance



from django.db.models import Field
from django.db.models.expressions import Func
from django.db.models.aggregates import Aggregate


class Annotation:

    def __init__(
            self,
            verbose_name: str,
            field: type[Field],
            function: type[Func] | type[Aggregate],
            help_text: str = '',
            **kwargs
    ):
        self.verbose_name = verbose_name or self
        self.field = field
        self.function = function
        self.help_text = help_text
        self.__dict__.update(kwargs)


class AnnotationModelMixin:

    @classmethod
    def get_annotations(cls) -> list[dict]:
        return list({'name': a, 'annotation': getattr(cls, a)} for a in filter(
            lambda a:
            not a.startswith('__')
            and isinstance(getattr(cls, a), Annotation),
            cls.__dict__
        ))


class AnnotationManagerMixin:

    def get_annotations(self):
        if not hasattr(self.model, 'get_annotations'):
            return {}
        return {
            annotation['name']: annotation['annotation'].function
            for annotation in self.model.get_annotations()
        }



# Generated by Django 5.0.7 on 2024-08-01 05:28

import accrete.annotation
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('accrete', '0002_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Tag',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=50, verbose_name='Name')),
                ('color', models.CharField(max_length=7, verbose_name='Color')),
                ('tag_group', models.CharField(max_length=255, verbose_name='Group')),
                ('tenant', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='accrete.tenant')),
            ],
            options={
                'verbose_name': 'Tag',
                'verbose_name_plural': 'Tags',
                'db_table': 'accrete_tag',
                'default_related_name': 'tags',
            },
            bases=(models.Model, accrete.annotation.AnnotationModelMixin),
        ),
        migrations.AddConstraint(
            model_name='tag',
            constraint=models.UniqueConstraint(fields=('tenant', 'name'), name='unique_tag_name_per_tenant'),
        ),
        migrations.AddConstraint(
            model_name='tag',
            constraint=models.UniqueConstraint(fields=('tenant', 'tag_group'), name='unique_tag_group_per_tenant'),
        ),
    ]



from django.urls import path
from . import views

app_name = 'user'

urlpatterns = [
    path('login/', views.LoginView.as_view(), name='login'),
    path('logout/', views.LogoutView.as_view(), name='logout'),
    path('detail/', views.user_detail, name='detail'),
    path('edit/', views.user_edit, name='edit'),
    path('edit/password', views.user_change_password, name='edit_password'),
    path('edit/email', views.user_change_email, name='edit_email')
]



from django.contrib import admin
from .models import Tag
from accrete.models import TenantManager


class TagAdmin(admin.ModelAdmin):
    model = Tag
    list_display = ('name', 'color', 'tag_group', 'tenant')
    search_fields = ('name', 'tag_group', 'tenant__name')


admin.site.register(Tag, TagAdmin)



import logging
import json
from urllib.parse import quote_plus
from uuid import uuid4
from dataclasses import dataclass, field
from functools import partial
from typing import Type, TypedDict
from django.utils.translation import gettext_lazy as _t
from django.db import models
from django.db.models import Model, QuerySet
from django.core.paginator import Paginator
from django.core import paginator, exceptions
from django.forms import Form, ModelForm
from accrete.annotation import Annotation
from accrete.utils.http import cast_param
from .elements import (
    ClientAction, BreadCrumb, TableField, TableFieldType, TableFieldAlignment,
    ClientActionGroup
)
from .filter import Filter
from accrete.utils import filter_from_querystring

_logger = logging.getLogger(__name__)

DEFAULT_PAGINATE_BY = 40


class DetailPagination(TypedDict):
    previous_object_url: str
    next_object_url: str
    current_object_idx: int
    total_objects: int


@dataclass(kw_only=True)
class BaseContext:

    extra: dict = field(default_factory=dict)

    def __post_init__(self):
        for key, value in self.extra.items():
            setattr(self, key, value)

    def dict(self):
        return {
            attr: getattr(self, attr, None) for attr
            in filter(lambda x: not x.startswith('_'), self.__dict__)
        }


@dataclass(kw_only=True)
class Context(BaseContext):

    title: str = ''
    breadcrumbs: list[BreadCrumb] = field(default_factory=list)
    actions: list[ClientAction | ClientActionGroup] = field(default_factory=list)


@dataclass
class TableContext(Context):

    object_label: str
    fields: list[TableField]
    list_page: paginator.Page
    pagination_param_str: str
    endless_scroll: bool
    filter: Filter
    object_param_str: str = field(default='', kw_only=True)
    field_selection: bool = True


def table_context_factory(model: Type[Model], params: dict, queryset=None, **kwargs) -> TableContext:
    if queryset is None:
        queryset = filter_from_querystring(model, params)
    page = list_page(
        queryset,
        cast_param(params, 'paginate_by', int, 40),
        cast_param(params, 'page', int, 1)
    )
    try:
        has_name_field = (
            model._meta.get_field('name').get_internal_type() == 'CharField'
        )
    except exceptions.FieldDoesNotExist:
        has_name_field = False

    ctx = partial(TableContext, **dict(
        title=str(model._meta.verbose_name_plural),
        object_label=str(model._meta.verbose_name),
        object_param_str=url_param_str(
            params,
            ['q', 'paginate_by', 'page', 'fields']
        ),
        fields=get_table_fields(
            cast_param(params, 'fields', json.loads, []), model
        ),
        list_page=page,
        pagination_param_str=url_param_str(
            params, ['q', 'paginate_by', 'fields']
        ),
        endless_scroll=True,
        filter=Filter(
            model,
            default_filter_term='name__icontains' if has_name_field else ''
        )
    ))
    return ctx(**kwargs)


@dataclass
class ListContext(Context):

    object_label: str
    object_param_str: str
    list_page: paginator.Page
    pagination_param_str: str
    filter: Filter = None
    endless_scroll: bool = True
    column_width: int = 12
    column_width_widescreen: int = None
    column_width_desktop: int = None
    column_height: int = 150
    column_height_unit: str = 'px'
    field_selection: bool = False

    def __post_init__(self):
        super().__post_init__()
        if self.column_width not in range(1, 13):
            _logger.warning(
                'ListContext parameter column_width should be in range 1 - 12'
            )
        if self.column_width_widescreen is None:
            self.column_width_widescreen = self.column_width
        if self.column_width_desktop is None:
            self.column_width_desktop = self.column_width


def list_context_factory(model: Type[Model], params: dict, queryset=None, paginate_by=None, **kwargs) -> ListContext:
    if queryset is None:
        queryset = filter_from_querystring(model, params)
    page = list_page(
        queryset,
        cast_param(params, 'paginate_by', int, paginate_by or 20),
        cast_param(params, 'page', int, 1)
    )
    try:
        has_name_field = (
            model._meta.get_field('name').get_internal_type() == 'CharField'
        )
    except exceptions.FieldDoesNotExist:
        has_name_field = False

    ctx = partial(ListContext, **dict(
        title=str(model._meta.verbose_name_plural),
        object_label=str(model._meta.verbose_name),
        object_param_str=url_param_str(
            params,
            ['q', 'paginate_by', 'page']
        ),
        list_page=page,
        pagination_param_str=url_param_str(
            params, ['q', 'paginate_by']
        ),
        endless_scroll=True,
        filter=Filter(
            model,
            default_filter_term='name__icontains' if has_name_field else ''
        ),
        column_width=4,
        column_height=150
    ))
    return ctx(**kwargs)


@dataclass
class DetailContext(Context):

    object: Model
    detail_page: DetailPagination = None
    pagination_param_str: str = ''


@dataclass
class FormContext(Context):

    form: Form | ModelForm
    form_id: str = 'form'
    form_method: str = 'post'
    form_action: str = ''


@dataclass
class ModalFormContext(BaseContext):

    title: str
    form: Form | ModelForm
    form_id: str = 'form'
    blocking: bool = True


@dataclass
class ModalContext(BaseContext):

    title: str
    blocking: bool = False
    modal_id: str = None

    def __post_init__(self):
        super().__post_init__()
        if not self.modal_id:
            self.modal_id = f'modal{str(uuid4())[:8]}'


def url_param_dict(get_params: dict) -> dict:
    return {
        key: f'&{key}={quote_plus(value)}'
        for key, value in get_params.items()
    }


def url_param_str(params: dict, extract: list[str] = None) -> str:
    """
    Return a URL Querystring from the given params dict.
    If extract is supplied, only specified keys will be used.
    """
    if extract:
        params = extract_url_params(params, extract)
    param_str = (
        "".join(str(value) for value in url_param_dict(params).values())
        .replace('&&', '&')
        .replace('?&', '?')
        .strip('?&')
    )
    return f'?{param_str}'


def extract_url_params(params: dict, keys: list[str]) -> dict:
    return {key: params[key] for key in keys if key in params}


def exclude_params(params: dict, keys: list[str]) -> dict:
    return {key: val for key, val in params.items() if key not in keys}


def list_page(
        queryset: QuerySet, paginate_by: int, page_number: int
) -> paginator.Page:
    pages = Paginator(queryset, per_page=paginate_by)
    return pages.page(
        page_number <= pages.num_pages and page_number or pages.num_pages
    )


def detail_page(queryset: QuerySet, obj: Model) -> dict:
    if not hasattr(obj, 'get_absolute_url'):
        _logger.warning(
            'Detail pagination disabled for models without the '
            'get_absolute_url attribute. Set paginate_by to 0 to '
            'deactivate pagination.'
        )
        return {}
    if obj not in queryset:
        url = obj.get_absolute_url()
        return {
            'previous_object_url': url,
            'next_object_url': url,
            'current_object_idx': 1,
            'total_objects': 1
        }

    idx = (*queryset,).index(obj)
    previous_object_url = (
        queryset[idx - 1] if idx - 1 >= 0 else queryset.last()
    ).get_absolute_url()
    next_object_url = (
        queryset[idx + 1] if idx + 1 <= queryset.count() - 1 else queryset.first()
    ).get_absolute_url()
    return {
        'previous_object_url': previous_object_url,
        'next_object_url': next_object_url,
        'current_object_idx': idx + 1,
        'total_objects': queryset.count()
    }


def form_actions(discard_url: str, form_id: str = 'form') -> list[ClientAction]:
    return [
        ClientAction(
            name=_t('Save'),
            submit=True,
            class_list=['is-success'],
            form_id=form_id,
        ),
        ClientAction(
            name=_t('Discard'),
            url=discard_url
        )
    ]


def get_table_fields(
        fields: list[str],
        model: type[Model],
        field_definition: dict[str | TableField] = None
) -> list[TableField]:

    def get_alignment(field_attr) -> TableFieldAlignment:
        number_field_types = (
            models.DecimalField, models.IntegerField, models.FloatField
        )
        if (
            isinstance(field_attr, number_field_types)
            or (
                isinstance(field_attr, Annotation)
                and field_attr.field in number_field_types
            )
        ):
            return TableFieldAlignment.RIGHT
        return TableFieldAlignment.LEFT

    def get_field_definition(f_name: str) -> TableField:
        if definition := field_definition.get(f_name):
            return definition
        try:
            return model.table_field_definition[f_name]
        except (AttributeError, KeyError):
            attr = getattr(model, f_name)
            if not isinstance(attr, Annotation) and hasattr(attr, 'field'):
                attr = attr.field
        is_relation = (
            hasattr(attr, 'related_model')
            and hasattr(attr.related_model, 'get_absolute_url')
        )
        field_type = TableFieldType.NONE
        if getattr(attr, 'choices', False):
            field_type = TableFieldType.CHOICE_DISPLAY
        return TableField(
            label=attr.verbose_name,
            name=f_name,
            field_type=field_type,
            alignment=get_alignment(attr),
            is_relation=is_relation
        )

    if field_definition is None:
        field_definition = {}
    table_fields = []
    for field_name in fields:
        try:
            table_fields.append(get_field_definition(field_name))
        except AttributeError as e:
            _logger.error(e)
            pass
    return table_fields



from django.db import models
from accrete.tenant import get_tenant
from accrete.annotation import AnnotationManagerMixin


class TenantManager(models.Manager, AnnotationManagerMixin):

    def get_queryset(self):
        queryset = super().get_queryset()
        if tenant := get_tenant():
            queryset = queryset.filter(tenant=tenant)
        return queryset.annotate(**self.get_annotations())

    def bulk_create(
            self,
            objs,
            batch_size=None,
            ignore_conflicts=False,
            update_conflicts=False,
            update_fields=None,
            unique_fields=None,
    ):
        tenant = get_tenant()
        if tenant is None and not all(obj.tenant_id for obj in objs):
            raise ValueError(
                'Tenant must be set for all objects when calling bulk_create'
            )
        for obj in objs:
            obj.tenant_id = tenant.pk
        return super().bulk_create(
            objs, batch_size=batch_size, ignore_conflicts=ignore_conflicts,
            update_conflicts=update_conflicts, update_fields=update_fields,
            unique_fields=unique_fields
        )



from django.utils.deprecation import MiddlewareMixin
from django.conf import settings
from django.http.response import HttpResponseRedirect, HttpResponsePermanentRedirect
from accrete.tenant import set_tenant, set_member

from .models import Tenant


class TenantMiddleware(MiddlewareMixin):
    @staticmethod
    def get_tenant_id_from_request(request):
        if tenant := request.POST.get('tenant'):
            tenant_id = tenant
        elif tenant := request.META.get('X_TENANT_ID'):
            tenant_id = tenant
        elif tenant_id := request.GET.get('tenant_id'):
            tenant_id = tenant_id
        elif tenant := request.COOKIES.get('tenant_id'):
            tenant_id = tenant
        else:
            tenant_id = None

        try:
            tenant_id = int(tenant_id)
        except (ValueError, TypeError):
            tenant_id = None

        return tenant_id

    def process_request(self, request):
        request.tenant = None
        request.member = None
        if request.path.startswith('/admin'):
            set_member(None)
            return

        if request.user.is_authenticated:
            memberships = request.user.memberships.filter(
                is_active=True,
                tenant__is_active=True
            )

            if request.user.is_staff:
                tenant_id = self.get_tenant_id_from_request(request)
                if not tenant_id:
                    tenant_id = memberships and memberships.first().tenant.pk
                request.tenant = tenant_id and Tenant.objects.filter(pk=tenant_id).first()
                request.member = request.tenant.members.filter(user=request.user).first()

            elif memberships and len(memberships) == 1:
                request.member = memberships.first()
                request.tenant = request.member.tenant
            elif memberships and len(memberships) > 1:
                tenant_id = self.get_tenant_id_from_request(request)
                if not tenant_id:
                    tenant_id = memberships.first().tenant.pk
                request.member = memberships.filter(tenant_id=tenant_id).first()
                request.tenant = request.member.tenant

        set_member(request.member)
        if request.user.is_staff:
            set_tenant(request.tenant)

        if request.POST and not request.POST.get('tenant') and request.tenant:
            request.POST = request.POST.copy()
            request.POST['tenant'] = request.tenant.pk

    @staticmethod
    def process_response(request, response):
        if request.tenant:
            response.set_cookie(
                'tenant_id',
                request.tenant.pk,
                samesite=settings.SESSION_COOKIE_SAMESITE,
                max_age=31536000
            )
        return response


class HtmxRedirectMiddleware(MiddlewareMixin):

    @staticmethod
    def process_response(request, response):
        is_htmx = request.headers.get('HX-Request', 'false') == 'true'
        if not is_htmx:
            return response
        is_redirect = isinstance(response, HttpResponseRedirect)
        is_permanent_redirect = isinstance(response, HttpResponsePermanentRedirect)
        if is_redirect or is_permanent_redirect:
            response.status_code = 200
            header = 'HX-Location' if is_redirect else 'HX-Redirect'
            response[header] = response['Location']
        return response



from django.urls import path

from . import views

app_name = 'user_registration'

urlpatterns = [
    path('register/', views.RegistrationView.as_view(), name='registration'),
    path('mail_resend/', views.ResendConfirmationMailView.as_view(), name='registration_resend_mail'),
    path('confirmation/<str:token>/', views.ConfirmRegistrationView.as_view(), name='confirmation')
]



from enum import Enum
from dataclasses import dataclass, field


class Icon(Enum):

    ADD = 'fas fa-plus'
    EDIT = 'fas fa-pencil'
    LIST = 'fas fa-list'
    OPEN_RELATED = 'icon-open-related'
    ENVELOPE = 'fas fa-envelope'
    IMAGE = 'fas fa-image'
    CLEAR = 'icon-clear'
    BACKSPACE = 'icon-backspace'
    FILTER = 'icon-filter'
    DELETE_FILTER = 'icon-delete-filter'
    SELECT = 'icon-select'
    TRASH_CAN = 'fa fa-trash-can'
    BOLT = 'fa fa-bolt'


class ActionMethod(Enum):

    HREF = 'href'
    GET = 'hx-get'
    POST = 'hx-post'
    PUT = 'hx-put'
    DELETE = 'hx-delete'


@dataclass
class ClientAction:

    name: str
    url: str = ''
    method: ActionMethod = ActionMethod.HREF
    attrs: list[tuple[str, str]] = field(default_factory=list)
    submit: bool = False
    form_id: str = 'form'
    class_list: list[str] = field(default_factory=list)
    icon: Icon | type[Enum] = None
    icon_only: bool = False

    def attrs_str(self):
        return ' '.join([f'{str(attr[0])}={str(attr[1])}' for attr in self.attrs])


@dataclass
class ClientActionGroup:

    name: str
    actions: list[ClientAction] = field(default_factory=list)
    icon: Icon | type[Enum] = None
    icon_only: bool = False


class TableFieldAlignment(Enum):

    LEFT = 'left'
    CENTER = 'center'
    RIGHT = 'right'


class TableFieldType(Enum):

    NONE = ''
    STRING = '_string'
    MONETARY = '_monetary'
    FLOAT = '_float'
    CHOICE_DISPLAY = '_choice_display'


@dataclass
class TableField:

    label: str
    name: str
    alignment: TableFieldAlignment | Enum = TableFieldAlignment.LEFT
    header_alignment: TableFieldAlignment | Enum = None
    header_info: str = None
    field_type: TableFieldType | Enum = TableFieldType.NONE
    prefix: str = ''
    suffix: str = ''
    truncate_after: int = 0
    template: str = None
    is_relation: bool = False


@dataclass
class BreadCrumb:

    name: str
    url: str



import logging

from django import forms
from django.core.validators import validate_email

from .models import SystemMail
from .tasks import run_mail_queue

_logger = logging.getLogger(__name__)


class SystemMailCreateForm(forms.ModelForm):

    class Meta:
        model = SystemMail
        fields = [
            'from_name',
            'to_addr',
            'subject',
            'body'
        ]

    def clean_to_addr(self):
        emails = self.cleaned_data['to_addr']
        for email in emails.split(','):
            validate_email(email)
        return emails

    def save(self, commit=True):
        super().save(commit)
        if commit:
            run_mail_queue.delay()
        return self.instance
